@@ -2,16 +2,24 @@
  * ml1.c
  *
  *
- * Part of TREE-PUZZLE 5.0 (June 2000)
+ * Part of TREE-PUZZLE 5.2 (July 2004)
  *
- * (c) 1999-2000 by Heiko A. Schmidt, Korbinian Strimmer,
+ * (c) 2003-2004 by Heiko A. Schmidt, Korbinian Strimmer, and Arndt von Haeseler
+ * (c) 1999-2003 by Heiko A. Schmidt, Korbinian Strimmer,
  *                  M. Vingron, and Arndt von Haeseler
  * (c) 1995-1999 by Korbinian Strimmer and Arndt von Haeseler
  *
  * All parts of the source except where indicated are distributed under
  * the GNU public licence.  See http://www.opensource.org for details.
+ *
+ * ($Id$)
+ *
  */
 
+#ifdef HAVE_CONFIG_H
+#  include <config.h>
+#endif
+
 
 /******************************************************************************/
 /* definitions and prototypes                                                 */
@@ -26,6 +34,12 @@
 #include <ctype.h>
 #include "util.h"
 #include "ml.h"
+/*epe*/
+#ifdef PARALLEL
+#	include <mpi.h>
+#	include "ppuzzle.h"
+#endif
+
 
 #define STDOUT     stdout
 #ifndef PARALLEL             /* because printf() runs significantly faster */
@@ -36,8 +50,14 @@
 #else
 #       define FPRINTF    fprintf
 #       define STDOUTFILE STDOUT,
+	void PP_Finalize();  /* prototype */
 #endif
 
+EXTERN int printrmatr_optn;  /* print rate matrix to screen */
+#if 0
+extern int PP_NumProcs; /*epe*/ /* might be included via ppuzzle.h ? */
+extern int PP_Myid; /*epe*/     /* might be included via ppuzzle.h ? */
+#endif
 
 /******************************************************************************/
 /* compacting sequence data information                                       */
@@ -75,10 +95,10 @@ void convfreq(dvector freqemp)
 			}
 		}
 	}
-}
+} /* convfreq */
 
 /* sort site patters of original input data */
-void radixsort(cmatrix seqchar, ivector ali, int maxspc, int maxsite,
+void tp_radixsort(cmatrix seqchar, ivector ali, int maxspc, int maxsite,
 	int *numptrn)
 {
 	int i, j, k, l, n, pass;
@@ -88,24 +108,44 @@ void radixsort(cmatrix seqchar, ivector ali, int maxspc, int maxsite,
 
 	awork = new_ivector(maxsite);
 	count = new_ivector(tpmradix+1);
-	for (i = 0; i < maxsite; i++)
+#ifndef USE_WINDOWS
+	for (i = 0; i < maxsite; i++)                                      /* init alignment column order 0, 1, 2, ... */
 		ali[i] = i;
-	for (pass = maxspc - 1; pass >= 0; pass--) {
-		for (j = 0; j < tpmradix+1; j++)
+#else
+	alimaxsize = aliwinend - aliwinstart + 1;
+	k = 0;
+	for (i = aliwinstart; i < aliwinend; i++) {
+		ali[k++] = i;
+#endif
+	for (pass = maxspc - 1; pass >= 0; pass--) {                       /* for every taxon */
+		for (j = 0; j < tpmradix+1; j++)                           /* reset count array */
 			count[j] = 0;
-		for (i = 0; i < maxsite; i++)
+#ifndef USE_WINDOWS
+		for (i = 0; i < maxsite; i++)                              /* counting bin size needed */
+#else
+		for (i = 0; i < alimaxsite; i++)                           /* counting bin size needed */
+#endif
 			count[(int) seqchar[pass][ali[i]]]++;
-		for (j = 1; j < tpmradix+1; j++)
+		for (j = 1; j < tpmradix+1; j++)                           /* converting bin sizes to start indices */
 			count[j] += count[j-1];
-		for (i = maxsite-1; i >= 0; i--)
+#ifndef USE_WINDOWS
+		for (i = maxsite-1; i >= 0; i--)                           /* produce consistently sorted order of alignment columns for current species to awork */
+#else
+		for (i = alimaxsite-1; i >= 0; i--)                        /* produce consistently sorted order of alignment columns for current species to awork */
+#endif
 			awork[ --count[(int) seqchar[pass][ali[i]]] ] = ali[i];
-		for (i = 0; i < maxsite; i++)
+		for (i = 0; i < maxsite; i++)                              /* set intermediate order */
 			ali[i] = awork[i];
 	}
 	free_ivector(awork);
 	free_ivector(count);
+
 	n = 1;
-	for (j = 1; j < maxsite; j++) {
+#ifndef USE_WINDOWS
+	for (j = 1; j < maxsite; j++) {                                    /* counting number of site pattern */
+#else
+	for (j = 1; j < alimaxsite; j++) {                                 /* counting number of site pattern */
+#endif
 		k = ali[j];
 		l = ali[j-1];
 		for (i = 0; i < maxspc; i++) {
@@ -116,7 +156,7 @@ void radixsort(cmatrix seqchar, ivector ali, int maxspc, int maxsite,
 		}
 	}
 	*numptrn = n;
-}
+} /* tp_radixsort */
 
 
 void condenceseq(cmatrix seqchar, ivector ali, cmatrix seqconint,
@@ -128,12 +168,20 @@ void condenceseq(cmatrix seqchar, ivector ali, cmatrix seqconint,
 
 	n = 0;
 	k = ali[n];
+#ifndef USE_WINDOWS
 	for (i = 0; i < maxspc; i++) {
+#else
+	for (i = aliwinstart - 1; i < aliwinend - 1; i++) {
+#endif
 		seqconint[i][n] = seqchar[i][k];
 	}
 	weight[n] = 1;
 	Alias[k] = 0;
+#ifndef USE_WINDOWS
 	for (j = 1; j < maxsite; j++) {
+#else
+	for (j = 1; j < alimaxsite; j++) {
+#endif
 		k = ali[j];
 		agree_flag = TRUE;
 		for (i = 0; i < maxspc; i++) {
@@ -157,10 +205,13 @@ void condenceseq(cmatrix seqchar, ivector ali, cmatrix seqconint,
 	n++;
 	if (numptrn != n) {
 		/* Problem in condenceseq */
-		FPRINTF(STDOUTFILE "\n\n\nHALT: PLEASE REPORT ERROR A TO DEVELOPERS\n\n\n");
+		fprintf(STDOUT, "\n\n\nHALT: PLEASE REPORT ERROR A TO DEVELOPERS\n\n\n");
+#		if PARALLEL
+			PP_Finalize();
+#		endif
 		exit(1);
 	}
-}
+} /* condenceseq */
 
 void countconstantsites(cmatrix seqpat, ivector weight, int maxspc, int numptrn,
 	int *numconst, int *numconstpat)
@@ -179,13 +230,16 @@ void countconstantsites(cmatrix seqpat, ivector weight, int maxspc, int numptrn,
 				break;
 			}
 		}
+
+		/* if constant pattern (patterns with wildcards are not */
+                /* constant)                                            */
 		if (character != tpmradix && constflag) {
 			(*numconst) = (*numconst) + weight[s];
 			(*numconstpat)++;
 			constpat[s] = TRUE;
 		}
 	}
-}
+} /* countconstantsites */
 
 /***************************** exported functions *****************************/
 
@@ -196,7 +250,7 @@ void evaluateseqs()
 
 	convfreq(Freqtpm); /* make all frequencies slightly different */
 	ali = new_ivector(Maxsite);
-	radixsort(Seqchar, ali, Maxspc, Maxsite, &Numptrn);
+	tp_radixsort(Seqchar, ali, Maxspc, Maxsite, &Numptrn);
 	Seqpat = new_cmatrix(Maxspc, Numptrn);
 	constpat = new_ivector(Numptrn);
 	Weight = new_ivector(Numptrn);
@@ -204,8 +258,12 @@ void evaluateseqs()
 	free_ivector(ali);
 	countconstantsites(Seqpat, Weight, Maxspc, Numptrn, &Numconst, &Numconstpat);
 	fracconstpat = (double) Numconstpat / (double) Numptrn;	
+#ifndef USE_WINDOWS
 	fracconst    = (double) Numconst / (double) Maxsite;	
-}
+#else
+	fracconst    = (double) Numconst / (double) alimaxsite;	
+#endif
+} /* evaluateseqs */
 
 
 /******************************************************************************/
@@ -260,7 +318,7 @@ void elmhes(dmatrix a, ivector ordr, int n)
 			}
 		}
 	}
-}
+} /* elmhes */
 
 
 void eltran(dmatrix a, dmatrix zz, ivector ordr, int n)
@@ -289,7 +347,7 @@ void eltran(dmatrix a, dmatrix zz, ivector ordr, int n)
 			zz[i - 1][m - 1] = 1.0;
 		}
 	}
-}
+} /* eltran */
 
 
 void mcdiv(double ar, double ai, double br, double bi,
@@ -306,7 +364,7 @@ void mcdiv(double ar, double ai, double br, double bi,
 	s = brs * brs + bis * bis;
 	*cr = (ars * brs + ais * bis) / s;
 	*ci = (ais * brs - ars * bis) / s;
-}
+} /* mcdiv */
 
 
 void hqr2(int n, int low, int hgh, dmatrix h,
@@ -353,7 +411,10 @@ void hqr2(int n, int low, int hgh, dmatrix h,
 				break;
 			if (itn == 0) {
 				/* all eigenvalues have not converged */
-				FPRINTF(STDOUTFILE "\n\n\nHALT: PLEASE REPORT ERROR B TO DEVELOPERS\n\n\n");
+				fprintf(STDOUT, "\n\n\nHALT: PLEASE REPORT ERROR B TO DEVELOPERS\n\n\n");
+#				if PARALLEL
+					PP_Finalize();
+#				endif
 				exit(1);
 			}
 			y = h[na - 1][na - 1];
@@ -679,7 +740,7 @@ void hqr2(int n, int low, int hgh, dmatrix h,
 		}
 	}
 	return;
-}
+} /* hqr2 */
 
 
 /* make rate matrix with 0.01 expected substitutions per unit time */
@@ -715,10 +776,43 @@ void onepamratematrix(dmatrix a)
 		}
 	}
 	free_dvector(m);
-}
+} /* onepamratematrix */
 
 
-void eigensystem(dvector eval, dmatrix evec)
+/* print rate matrix to fp for testing purposes */
+void printratematrix(FILE *fp, dmatrix a, int do_multpi)
+{
+	int i, j;
+
+	double sum = 0.0;
+	double rowsum;
+
+	/* fprintf(fp, "\n"); */
+	for (i = 0; i < tpmradix; i++)
+	{
+		rowsum = 0.0;
+		fprintf(fp, "%-10s ",int2code(i));
+		for (j = 0; j < tpmradix; j++)
+		{
+			if (! do_multpi) {
+				fprintf(fp, "%10.5f", a[i][j]);
+				if (i != j) rowsum += a[i][j];
+			} else {
+				fprintf(fp, "%10.5f", a[i][j] * Freqtpm[i]);
+				if (i != j) rowsum += a[i][j] * Freqtpm[i];
+				/* if (i != j) rowsum += Freqtpm[i] * a[i][j]; */
+			}
+		}
+		sum += rowsum;
+		fprintf(fp, " row=%10.5f\n",rowsum);
+	}
+	fprintf(fp, "off-diagonal sum = %10.5f\n", sum);
+} /* printratematrix */
+
+
+void eigensystem(
+	dvector eval, 	/* vector of Eigen-values */
+	dmatrix evec)	/* vector of Eigen-vectors*/
 {
 	dvector evali, forg;
 	dmatrix a, b;
@@ -734,8 +828,20 @@ void eigensystem(dvector eval, dmatrix evec)
 	b = new_dmatrix(tpmradix,tpmradix);
 
 	rtfdata(a, forg); /* get relative transition matrix and frequencies */
+	if (printrmatr_optn) {
+		fprintf(stdout, "Current rate matrix R :\n");
+		printratematrix(stdout, a, FALSE); /* print rate matrix */
+		printratematrix(stdout, a, TRUE); /* print rate matrix */
+		fprintf(stdout, "\n");
+	}
 	
 	onepamratematrix(a); /* make 1 PAM rate matrix */
+	if (printrmatr_optn) {
+		fprintf(stdout, "Current transition matrix Q (1 PAM normalized):\n"); /* print rate matrix */
+		printratematrix(stdout, a, FALSE); /* print rate matrix multiplied with pi */
+		printratematrix(stdout, a, TRUE); /* print rate matrix multiplied with pi */
+	}
+	
 	
 	/* copy a to b */
 	for (i = 0; i < tpmradix; i++)
@@ -757,14 +863,14 @@ void eigensystem(dvector eval, dmatrix evec)
 		}
 	}
 	if (error)
-		FPRINTF(STDOUTFILE "\nWARNING: Eigensystem doesn't satisfy eigenvalue equation!\n");
+		fprintf(STDOUT, "\nWARNING: Eigensystem doesn't satisfy eigenvalue equation!\n");
 
 	free_ivector(ordr);
 	free_dvector(evali);
 	free_dvector(forg);
 	free_dmatrix(a);
 	free_dmatrix(b);
-}
+} /* eigensystem */
 
 
 void luinverse(dmatrix inmat, dmatrix imtrx, int size)
@@ -785,7 +891,7 @@ void luinverse(dmatrix inmat, dmatrix imtrx, int size)
 		for (j = 0; j < tpmradix; j++)
 			omtrx[i][j] = inmat[i][j];
 	
-	wk = (double *) malloc((unsigned)size * sizeof(double));
+	wk = (double *) calloc((size_t)size, sizeof(double));
 	aw = 1.0;
 	for (i = 0; i < size; i++) {
 		maxb = 0.0;
@@ -795,7 +901,10 @@ void luinverse(dmatrix inmat, dmatrix imtrx, int size)
 		}
 		if (maxb == 0.0) {
 			/* Singular matrix */
-			FPRINTF(STDOUTFILE "\n\n\nHALT: PLEASE REPORT ERROR C TO DEVELOPERS\n\n\n");
+			fprintf(STDOUT, "\n\n\nHALT: PLEASE REPORT ERROR C TO DEVELOPERS\n\n\n");
+#			if PARALLEL
+				PP_Finalize();
+#			endif
 			exit(1);
 		}
 		wk[i] = 1.0 / maxb;
@@ -866,7 +975,7 @@ void luinverse(dmatrix inmat, dmatrix imtrx, int size)
 	wk = NULL;
 	free_ivector(index);
 	free_dmatrix(omtrx);
-}
+} /* luinverse */
 
 
 void checkevector(dmatrix evec, dmatrix ivec, int nn)
@@ -899,10 +1008,10 @@ void checkevector(dmatrix evec, dmatrix ivec, int nn)
 		}
 	}
 	if (error) {
-		FPRINTF(STDOUTFILE "\nWARNING: Inversion of eigenvector matrix not perfect!\n");
+		fprintf(STDOUT, "\nWARNING: Inversion of eigenvector matrix not perfect!\n");
 	}
 	free_dmatrix(matx);
-}
+} /* checkevector */
 
 
 /***************************** exported functions *****************************/
@@ -914,7 +1023,7 @@ void tranprobmat()
 	eigensystem(Eval, Evec); /* eigensystem of 1 PAM rate matrix */
 	luinverse(Evec, Ievc, tpmradix); /* inverse eigenvectors are in Ievc */
 	checkevector(Evec, Ievc, tpmradix); /* check whether inversion was OK */
-}
+} /* tranprobmat */
 
 
 /* compute P(t) */
@@ -937,7 +1046,7 @@ void tprobmtrx(double arc, dmatrix tpr)
 			tpr[i][j] = fabs(temp);
 		}
 	}
-}
+} /* tprobmtrx */
 
 
 /******************************************************************************/
@@ -993,7 +1102,7 @@ double comptotloglkl(dmatrix cdl)
 	}
 	
 	return loglkl;
-}
+} /* comptotloglkl */
 
 
 /* computes the site log-likelihoods 
@@ -1040,7 +1149,82 @@ void allsitelkl(dmatrix cdl, dvector aslkl)
 		
 		}
 	}
-}
+} /* allsitelkl */
+
+
+/* YYYY */
+/* computes the site log-likelihoods
+   input: likelihoods for each site and non-zero rate
+   output: log-likelihood for each site */
+void writesitelklvectorbin(FILE *outf, dvector aslkl)
+{
+	int s;
+	int dsize;
+
+	dsize = sizeof(double);
+
+	for (s = 0; s < Maxsite; s++) {
+		fprintf(stdout, "%4d (%4d) %.5f \n", s, Alias[s], aslkl[Alias[s]]);
+		if( 1 != fwrite( &(aslkl[Alias[s]]), dsize, 1, outf) ) {
+			fprintf(stderr, "ERROR writing log-likelihood of site %d to file!!!\n", Alias[s]);
+#			if PARALLEL
+				PP_Finalize();
+#			endif
+			exit (1);
+		}
+	}
+
+
+} /* writesitelklvectorbin */
+
+/* outputs a name and its log-likelihoods to outf */
+void writesitelklvectorphy(FILE *outf, char *name, int numsite, dvector aslkl)
+{
+	int s;
+	int dsize;
+
+	dsize = sizeof(double);
+
+	fprintf(outf, "%s", name);
+	for (s = 0; s < Maxsite; s++) {
+		if( 0 == fprintf(outf, " %.5f", aslkl[Alias[s]])) {
+			fprintf(stderr, "ERROR writing log-likelihood of site %d for '%s' to file!!!\n", Alias[s], name);
+#			if PARALLEL
+				PP_Finalize();
+#			endif
+			exit (1);
+		}
+	}
+	fprintf(outf, "\n");
+} /* writesitelklvectorphy */
+
+
+/* outputs a name and its log-likelihoods to outf */
+void writesitelklmatrixphy(FILE *outf, int numut, int numsite, dmatrix allslkl, dmatrix allslklc, int clock)
+{
+	int t;
+	char tmpname[20];
+
+	if (clock) 
+		fprintf(outf, " %d %d \n", 2 * numut, numsite);
+	else
+		fprintf(outf, " %d %d \n", numut, numsite);
+
+	for (t = 0; t < numut; t++) {
+		sprintf(tmpname, "tr%-7d  ", t+1);
+		writesitelklvectorphy(outf, tmpname, numsite, allslkl[t]);
+	}
+
+	if (clock) {
+		for (t = 0; t < numut; t++) {
+			sprintf(tmpname, "ctr%-7d ", t+1);
+			writesitelklvectorphy(outf, tmpname, numsite, allslklc[t]);
+		}
+	}
+
+} /* writesitelklmatrixphy */
+
+
 
 
 /***************************** internal functions *****************************/
@@ -1108,21 +1292,22 @@ double pairlkl(double arc)
 	
 	/* return negative log-likelihood as we use a minimizing procedure */
 	return -loglkl;
-}
+} /* pairlkl */
 
 
 /***************************** exported functions *****************************/
 
 
 /* maximum likelihood distance between sequence i and j */
-double mldistance(int i, int j)
+double mldistance(int i, int j, double init)
 {
 	double dist, fx, f2x;
-	
+
 	if (i == j) return 0.0;
-	
-	/* use old distance as start value */	
-	dist = Distanmat[i][j];
+
+	/* use old distance as start value */
+/*epe	dist = Distanmat[i][j];*/
+	dist = init;
 
 	if (dist == 0.0) return 0.0;
 
@@ -1132,10 +1317,10 @@ double mldistance(int i, int j)
 	if (dist <= MINARC) dist = MINARC+1.0;
 	if (dist >= MAXARC) dist = MAXARC-1.0;
 	
- 	dist = onedimenmin(MINARC, dist, MAXARC, pairlkl, EPSILON, &fx, &f2x);
+ 	dist = onedimenmin(MINARC, dist, MAXARC, pairlkl, EPSILON_BRANCH, &fx, &f2x);
   	
 	return dist;
-}
+} /* mldistance */
 
 
 /* initialize distance matrix */
@@ -1166,7 +1351,11 @@ void initdistan()
 				/* use generalized JC correction to get first estimate
 				   (for the SH model the observed distance is used) */
 				/* observed distance */
+#ifndef USE_WINDOWS
 				obs = (double) diff / (double) Maxsite;
+#else
+				obs = (double) diff / (double) alimaxsite;
+#endif
 				temp = 1.0 - (double) obs*tpmradix/(tpmradix-1.0);
 				if (temp > 0.0 && !(data_optn == 0 && SH_optn))
 					/* use JC corrected distance */
@@ -1180,19 +1369,126 @@ void initdistan()
 			Distanmat[j][i] = Distanmat[i][j];
 		}
 	}
-}
+} /* initdistan */
 
 /* compute distance matrix */
+/*epe*/
+#if PARALLEL
+void computedistan()
+{
+	int i, slaveno, loop;
+	int xr, yr, xs=0, ys=1, numstartjobs=0, dummychunk=0;
+	dvector startval, dummyval=new_dvector(1);
+	ivector startlen = new_ivector(PP_NumProcs), startpoint = new_ivector(PP_NumProcs);
+	imatrix coords = new_imatrix(PP_NumProcs, 3);
+	schedtype sched;
+	MPI_Status stat;
+
+	loop = .5 * Maxspc * (Maxspc-1);
+	initsched(&sched, loop, PP_NumProcs-1, 1);
+	if (PP_NumProcs < loop) loop = PP_NumProcs-1;
+
+	startlen[0] = 0;
+	startpoint[0] = 0;
+
+	/*Send first min(#numprocs, #matrix entries) matrix entries - one to each processor.
+	If #numprocs > #matrix entries, send all matrix entries; some processors remain idle.*/
+	for (slaveno = 1; slaveno < PP_NumProcs; slaveno++)
+	{
+
+		if (slaveno<=loop)
+		{
+			unsigned int chunks = SCHEDALG_PARAM_EST(&sched);
+
+			startlen[slaveno] = chunks;
+			startpoint[slaveno] = numstartjobs;
+			numstartjobs += chunks;
+
+			coords[slaveno][0] = xs;
+			coords[slaveno][1] = ys;
+			coords[slaveno][2] = chunks;
+
+			PP_NextCoord(&xs, &ys, chunks);
+		}
+		else startlen[slaveno] = 0;
+	}
+	startval = new_dvector(numstartjobs);
+	PP_DMToVector(0, 1, numstartjobs, startval);
+	MPI_Scatterv(startval, startlen, startpoint, MPI_DOUBLE,
+		dummyval, dummychunk, MPI_DOUBLE, PP_MyMaster, PP_Comm);
+	free_dvector(startval);
+
+	/*The remaining matrix entries (if there are any) are sent to those processors that are
+	returning results.*/
+	while (xs<Maxspc-1)
+	{
+		dvector result, initval;
+		unsigned int chunkr, chunks = SCHEDALG_PARAM_EST(&sched);
+
+		MPI_Probe(MPI_ANY_SOURCE, PP_MLDISTANCE, PP_Comm, &stat);
+		slaveno = stat.MPI_SOURCE;
+
+		xr = coords[slaveno][0];
+		yr = coords[slaveno][1];
+		chunkr = coords[slaveno][2];
+
+		result = new_dvector(chunkr);
+
+		MPI_Recv(result, chunkr, MPI_DOUBLE, slaveno, PP_MLDISTANCE, PP_Comm, &stat);
+
+		PP_VectorToDM(xr, yr, chunkr, result);
+
+		free_dvector(result);
+
+		coords[slaveno][0] = xs;
+		coords[slaveno][1] = ys;
+		coords[slaveno][2] = chunks;
+
+		initval = new_dvector(chunks+2);
+		PP_DMToVector(xs, ys, chunks, initval);
+		initval[chunks] = (double) -xs;
+		initval[chunks+1] = (double) ys;
+
+		MPI_Send(initval, chunks+2, MPI_DOUBLE, slaveno, PP_MLDISTANCE, PP_Comm);
+
+		PP_NextCoord(&xs, &ys, chunks);
+	}
+
+	/*min(#numprocs, #matrix entries) results have to be received.*/
+	for (i = 1; i <= loop; i++)
+	{
+		double* result;
+		unsigned int chunkr;
+
+		MPI_Probe(MPI_ANY_SOURCE, PP_MLDISTANCE, PP_Comm, &stat);
+		slaveno = stat.MPI_SOURCE;
+
+		xr = coords[slaveno][0];
+		yr = coords[slaveno][1];
+		chunkr = coords[slaveno][2];
+
+		result = new_dvector(chunkr);
+
+		MPI_Recv(result, chunkr, MPI_DOUBLE, slaveno, PP_MLDISTANCE, PP_Comm, &stat);
+
+		PP_VectorToDM(xr, yr, chunkr, result);
+	}
+	return;
+} /* computedistan - parallel */
+
+#else /* not PARALLEL */
+
 void computedistan()
 {
 	int i, j;
 
 	for (i = 0; i < Maxspc; i++)
 		for (j = i; j < Maxspc; j++) {
-			Distanmat[i][j] = mldistance(i, j);
+			Distanmat[i][j] = mldistance(i, j, Distanmat[i][j]);
 			Distanmat[j][i] = Distanmat[i][j];
 		}
-}
+} /* computedistan */
+#endif /* not PARALLEL */
 
 
 /******************************************************************************/
@@ -1220,7 +1516,7 @@ void productpartials(Node *op)
 				for (j = 0; j < tpmradix; j++)
 					opc[r][i][j] *= cpc[r][i][j];
 	}
-}
+} /* productpartials */
 
 
 /* compute internal partial likelihoods */
@@ -1252,7 +1548,7 @@ void partialsinternal(Node *op)
 			}
 		}
 	}
-}
+} /* partialsinternal */
 
 
 /* compute external partial likelihoods */
@@ -1285,7 +1581,7 @@ void partialsexternal(Node *op)
 			}
 		}
 	}
-}
+} /* partialsexternal */
 
 
 /* compute all partial likelihoods */
@@ -1306,7 +1602,7 @@ void initpartials(Tree *tr)
 			}
 		}
 	} while (cp != rp);
-}
+} /* initpartials */
 
 
 /* compute log-likelihood given internal branch with length arc
@@ -1338,7 +1634,7 @@ double intlkl(double arc)
 	Ctree->lklhd = comptotloglkl(cdl);
 
 	return -(Ctree->lklhd); /* we use a minimizing procedure */
-}
+} /* intlkl */
 
 
 /* optimize internal branch */
@@ -1351,7 +1647,7 @@ void optinternalbranch(Node *op)
 	arc = op->length; /* nonclocklike branch lengths */
 	if (arc <= MINARC) arc = MINARC+1.0;
 	if (arc >= MAXARC) arc = MAXARC-1.0;
-	arc = onedimenmin(MINARC, arc, MAXARC, intlkl, EPSILON, &fx, &f2x);
+	arc = onedimenmin(MINARC, arc, MAXARC, intlkl, EPSILON_BRANCH, &fx, &f2x);
 	op->kinp->length = arc;
 	op->length = arc;
 	
@@ -1361,7 +1657,7 @@ void optinternalbranch(Node *op)
 		op->varlen = 1.0/f2x;
 	else
 		op->varlen = MAXARC*MAXARC;
-}
+} /* optinternalbranch */
 
 
 /* compute log-likelihood given external branch with length arc
@@ -1396,7 +1692,7 @@ double extlkl(double arc)
 	Ctree->lklhd = comptotloglkl(cdl);
 
 	return -(Ctree->lklhd); /* we use a minimizing procedure */
-}
+} /* extlkl */
 
 /* optimize external branch */
 void optexternalbranch(Node *op)
@@ -1408,7 +1704,7 @@ void optexternalbranch(Node *op)
 	arc = op->length; /* nonclocklike branch lengths */
 	if (arc <= MINARC) arc = MINARC+1.0;
 	if (arc >= MAXARC) arc = MAXARC-1.0;
-	arc = onedimenmin(MINARC, arc, MAXARC, extlkl, EPSILON, &fx, &f2x);
+	arc = onedimenmin(MINARC, arc, MAXARC, extlkl, EPSILON_BRANCH, &fx, &f2x);
 	op->kinp->length = arc;
 	op->length = arc;
 	
@@ -1418,7 +1714,7 @@ void optexternalbranch(Node *op)
 		op->varlen = 1.0/f2x;
 	else
 		op->varlen = MAXARC*MAXARC;
-}
+} /* optexternalbranch */
 
 
 /* finish likelihoods for each rate and site */
@@ -1447,7 +1743,7 @@ void finishlkl(Node *op)
 			cdl[r][k] = sumlk;
 		}
 	}
-}
+} /* finishlkl */
 
 
 /***************************** exported functions *****************************/
@@ -1476,7 +1772,7 @@ double optlkl(Tree *tr)
 				lendiff = cp->length;
 				optexternalbranch(cp);
 				lendiff = fabs(lendiff - cp->length);
-				if (lendiff < EPSILON) nconv++;
+				if (lendiff < EPSILON_BRANCH) nconv++;
 				else nconv = 0;				
 				
 				partialsexternal(cp);
@@ -1488,11 +1784,11 @@ double optlkl(Tree *tr)
 					lendiff = cp->length;
 					optinternalbranch(cp);
 					lendiff = fabs(lendiff - cp->length);
-					if (lendiff < EPSILON) nconv++;
+					if (lendiff < EPSILON_BRANCH) nconv++;
 					else nconv = 0;
 					
 					/* eventually compute likelihoods for each site */
-					if ((cp->number == Numibrnch-1 && lendiff < EPSILON) ||
+					if ((cp->number == Numibrnch-1 && lendiff < EPSILON_BRANCH) ||
 					Numit == MAXIT-1) finishlkl(cp);
 
 					partialsinternal(cp);
@@ -1507,7 +1803,7 @@ double optlkl(Tree *tr)
 
 	/* compute total log-likelihood for current tree */
 	return comptotloglkl(tr->condlkl);
-}
+} /* optlkl */
 
 
 /* compute likelihood of tree for given branch lengths */
@@ -1537,7 +1833,7 @@ double treelkl(Tree *tr)
 	
 	/* return total log-likelihood for current tree */
 	return comptotloglkl(cdl);
-}
+} /* treelkl */
 
 
 /******************************************************************************/
@@ -1568,7 +1864,10 @@ void luequation(dmatrix amat, dvector yvec, int size)
 		}
 		if (maxb == 0.0) {
 			/* Singular matrix */
-			FPRINTF(STDOUTFILE "\n\n\nHALT: PLEASE REPORT ERROR D TO DEVELOPERS\n\n\n");
+			fprintf(STDOUT, "\n\n\nHALT: PLEASE REPORT ERROR D TO DEVELOPERS\n\n\n");
+#			if PARALLEL
+				PP_Finalize();
+#			endif
 			exit(1);
 		}
 		wk[i] = 1.0 / maxb;
@@ -1630,12 +1929,285 @@ void luequation(dmatrix amat, dvector yvec, int size)
 	}
 	free_ivector(index);
 	free_dvector(wk);
-}
+} /* luequation */
 
 
 /* least square estimation of branch lengths
-   used for the approximate ML and as starting point 
+   used for the approximate ML and as starting point
    in the calculation of the exact value of the ML */
+/*epe*/
+#if PARALLEL
+void lslength_par(Tree *tr, dvector distanvec, int numspc, int numibrnch, dvector Brnlength)
+{
+	int i, i1, j, j1, j2, k, numbrnch, numpair, initmsg[2];
+	unsigned long int isum;
+	double sum, leng, alllen, rss;
+	ivector pths, jobsize = new_ivector(PP_NumProcs), atamt_tmp_coord = new_ivector(PP_NumProcs);
+	dvector my_atamt_tmp, atamt_tmp;
+	ivector IPaths;
+	cmatrix atmt;
+	dmatrix atamt;
+	Node **ebp, **ibp;
+	int numjobs, my_jobsize, my_x=0, my_y=0;
+
+	if (PP_Myid != PP_MyMaster)
+	{
+		fprintf(STDOUT, "ERROR: Slave [%d] in parallel lslength()!\n", PP_Myid);
+		return;
+	}
+
+	numbrnch = numspc + numibrnch;
+	numpair = (numspc * (numspc - 1)) / 2;
+	atmt = new_cmatrix(numbrnch, numpair);
+	atamt = new_dmatrix(numbrnch, numbrnch);
+
+	initmsg[0] = numspc; initmsg[1] = numibrnch;
+	for (i=1; i<PP_NumProcs; i++) MPI_Send(initmsg, 2, MPI_INT, i, PP_LSLENGTH, PP_Comm);
+
+	ibp = tr->ibrnchp;
+	IPaths = new_ivector(numibrnch*numspc);
+	for (i=0; i<numibrnch; i++) for (j=0; j<numspc; j++)
+		IPaths[i*numspc+j] = ibp[i]->paths[j];
+
+	MPI_Bcast(IPaths, numibrnch*numspc, MPI_INT, PP_MyMaster, PP_Comm);
+	free_ivector(IPaths);
+
+	numjobs = (numbrnch * (numbrnch+1))/2;
+	my_jobsize = numjobs/PP_NumProcs;
+	if (numjobs%PP_NumProcs>0) my_jobsize++;
+	jobsize[0] = my_jobsize;
+	for (i=1; i<PP_NumProcs; i++)
+	{
+		jobsize[i] = numjobs/PP_NumProcs;
+		if (i < numjobs%PP_NumProcs) jobsize[i]++;
+		atamt_tmp_coord[i] = atamt_tmp_coord[i-1]+jobsize[i-1];
+	}
+
+	for (i = 0; i < numspc; i++) {
+		for (j1 = 1, j = 0; j1 < numspc; j1++) {
+			if (j1 == i) {
+				for (j2 = 0; j2 < j1; j2++, j++) {
+					atmt[i][j] = 1;
+				}
+			} else {
+				for (j2 = 0; j2 < j1; j2++, j++) {
+					if (j2 == i)
+						atmt[i][j] = 1;
+					else
+						atmt[i][j] = 0;
+				}
+			}
+		}
+	}
+	for (i1 = 0, i = numspc; i1 < numibrnch; i1++, i++) {
+		pths = ibp[i1]->paths;
+		for (j1 = 1, j = 0; j1 < numspc; j1++) {
+			for (j2 = 0; j2 < j1; j2++, j++) {
+				if (pths[j1] != pths[j2])
+					atmt[i][j] = 1;
+				else
+					atmt[i][j] = 0;
+			}
+		}
+	}
+	if (PP_Myid == PP_MyMaster)
+	{
+		my_atamt_tmp = new_dvector(my_jobsize);
+#ifdef VT
+		VT_begin(1020);
+#endif
+		for (i = 0; i < my_jobsize; i++) {
+			for (k = 0, isum = 0; k < numpair; k++)
+				if (atmt[my_x][k] && atmt[my_y][k]) isum ++;
+			my_atamt_tmp[i] = (double) isum;
+			PP_NextCoord_atamt(&my_x, &my_y, 1, numbrnch);
+		}
+#ifdef VT
+		VT_end(1020);
+#endif
+		atamt_tmp = new_dvector(numjobs);
+		MPI_Gatherv(my_atamt_tmp, my_jobsize, MPI_DOUBLE,
+			atamt_tmp, jobsize, atamt_tmp_coord, MPI_DOUBLE, PP_MyMaster, PP_Comm);
+		free_dvector(my_atamt_tmp);
+		k=0;
+		for (i = 0; i < numbrnch; i++) {
+			for (j = 0; j <= i; j++) {
+				atamt[i][j] = atamt_tmp[k];
+				atamt[j][i] = atamt_tmp[k];
+				k++;
+			}
+		}
+		free_dvector(atamt_tmp);
+	}
+	else
+	{
+#ifdef VT
+		VT_begin(1020);
+#endif
+		for (i = 0; i < numbrnch; i++) {
+			for (j = 0; j <= i; j++) {
+				for (k = 0, isum = 0; k < numpair; k++)
+					if (atmt[i][k] && atmt[j][k]) isum ++;
+				atamt[i][j] = (double) isum;
+				atamt[j][i] = (double) isum;
+			}
+		}
+	}
+	for (i = 0; i < numbrnch; i++) {
+		for (k = 0, sum = 0.0; k < numpair; k++)
+			if (atmt[i][k]) sum += distanvec[k];
+		Brnlength[i] = sum;
+	}
+	luequation(atamt, Brnlength, numbrnch);
+	for (i = 0, rss = 0.0; i < numpair; i++) {
+		sum = distanvec[i];
+		for (j = 0; j < numbrnch; j++) {
+			if (atmt[j][i] && Brnlength[j] > 0.0)
+				sum -= Brnlength[j];
+		}
+		rss += sum * sum;
+	}
+	tr->rssleast = sqrt(rss);
+	alllen = 0.0;
+	ebp = tr->ebrnchp;
+	for (i = 0; i < numspc; i++) {
+		leng = Brnlength[i];
+		alllen += leng;
+		if (leng < MINARC) leng = MINARC;
+		if (leng > MAXARC) leng = MAXARC;
+		if (clockmode) { /* clock */
+			ebp[i]->lengthc = leng;
+			ebp[i]->kinp->lengthc = leng;
+		} else { /* no clock */
+			ebp[i]->length = leng;
+			ebp[i]->kinp->length = leng;
+		}
+		Brnlength[i] = leng;
+	}
+	for (i = 0, j = numspc; i < numibrnch; i++, j++) {
+		leng = Brnlength[j];
+		alllen += leng;
+		if (leng < MINARC) leng = MINARC;
+		if (leng > MAXARC) leng = MAXARC;
+		if (clockmode) { /* clock */
+			ibp[i]->lengthc = leng;
+			ibp[i]->kinp->lengthc = leng;
+		} else { /* no clock */
+			ibp[i]->length = leng;
+			ibp[i]->kinp->length = leng;
+		}
+		Brnlength[j] = leng;
+	}
+	free_cmatrix(atmt);
+	free_dmatrix(atamt);
+} /* lslength_par */
+
+#endif /* PARALLEL */
+
+/* Ekki's sped-up version of lslength */
+void lslength(Tree *tr, dvector distanvec, int numspc, int numibrnch, dvector Brnlength)
+{
+	int i, i1, j, j1, j2, k, numbrnch, numpair;
+	unsigned long int isum;
+	double sum, leng, alllen, rss;
+	ivector pths;
+	cmatrix atmt;
+	dmatrix atamt;
+	Node **ebp, **ibp;
+
+	numbrnch = numspc + numibrnch;
+	numpair = (numspc * (numspc - 1)) / 2;
+	atmt = new_cmatrix(numbrnch, numpair);
+	atamt = new_dmatrix(numbrnch, numbrnch);
+	ebp = tr->ebrnchp;
+	ibp = tr->ibrnchp;
+
+	for (i = 0; i < numspc; i++) {
+		for (j1 = 1, j = 0; j1 < numspc; j1++) {
+			if (j1 == i) {
+				for (j2 = 0; j2 < j1; j2++, j++) {
+					atmt[i][j] = 1;
+				}
+			} else {
+				for (j2 = 0; j2 < j1; j2++, j++) {
+					if (j2 == i)
+						atmt[i][j] = 1;
+					else
+						atmt[i][j] = 0;
+				}
+			}
+		}
+	}
+	for (i1 = 0, i = numspc; i1 < numibrnch; i1++, i++) {
+		pths = ibp[i1]->paths;
+		for (j1 = 1, j = 0; j1 < numspc; j1++) {
+			for (j2 = 0; j2 < j1; j2++, j++) {
+				if (pths[j1] != pths[j2])
+					atmt[i][j] = 1;
+				else
+					atmt[i][j] = 0;
+			}
+		}
+	}
+	for (i = 0; i < numbrnch; i++) {
+		for (j = 0; j <= i; j++) {
+			for (k = 0, isum = 0; k < numpair; k++)
+				if (atmt[i][k] && atmt[j][k]) isum ++;
+			atamt[i][j] = (double) isum;
+			atamt[j][i] = (double) isum;
+		}
+	}
+	for (i = 0; i < numbrnch; i++) {
+		for (k = 0, sum = 0.0; k < numpair; k++)
+			sum += (double) atmt[i][k] * distanvec[k];
+		Brnlength[i] = sum;
+	}
+	luequation(atamt, Brnlength, numbrnch);
+	for (i = 0, rss = 0.0; i < numpair; i++) {
+		sum = distanvec[i];
+		for (j = 0; j < numbrnch; j++) {
+			if (atmt[j][i] && Brnlength[j] > 0.0)
+				sum -= Brnlength[j];
+		}
+		rss += sum * sum;
+	}
+	tr->rssleast = sqrt(rss);
+	alllen = 0.0;
+	for (i = 0; i < numspc; i++) {
+		leng = Brnlength[i];
+		alllen += leng;
+		if (leng < MINARC) leng = MINARC;
+		if (leng > MAXARC) leng = MAXARC;
+		if (clockmode) { /* clock */
+			ebp[i]->lengthc = leng;
+			ebp[i]->kinp->lengthc = leng;
+		} else { /* no clock */
+			ebp[i]->length = leng;
+			ebp[i]->kinp->length = leng;
+		}
+		Brnlength[i] = leng;
+	}
+	for (i = 0, j = numspc; i < numibrnch; i++, j++) {
+		leng = Brnlength[j];
+		alllen += leng;
+		if (leng < MINARC) leng = MINARC;
+		if (leng > MAXARC) leng = MAXARC;
+		if (clockmode) { /* clock */
+			ibp[i]->lengthc = leng;
+			ibp[i]->kinp->lengthc = leng;
+		} else { /* no clock */
+			ibp[i]->length = leng;
+			ibp[i]->kinp->length = leng;
+		}
+		Brnlength[j] = leng;
+	}
+	free_cmatrix(atmt);
+	free_dmatrix(atamt);
+} /* lslength */
+
+
+#if 0
+/* old slow least square fit */
 void lslength(Tree *tr, dvector distanvec, int numspc, int numibrnch, dvector Brnlength)
 {
 	int i, i1, j, j1, j2, k, numbrnch, numpair;
@@ -1731,4 +2303,67 @@ void lslength(Tree *tr, dvector distanvec, int numspc, int numibrnch, dvector Br
 	}
 	free_dmatrix(atmt);
 	free_dmatrix(atamt);
-}
+} /* lslength */
+#endif
+
+/* set branch lengths to the externally given lengths  (-usebranchlen) */
+void setextlength(Tree *tr, int numspc, int numibrnch, dvector Brnlength)
+{
+	int i, j, numbrnch;
+	double leng, alllen;
+	Node **ebp, **ibp;
+
+	numbrnch = numspc + numibrnch;
+	ebp = tr->ebrnchp;
+	ibp = tr->ibrnchp;
+
+	alllen = 0.0;
+	for (i = 0; i < numspc; i++) {
+		if (ebp[i]->lengthset) {
+			leng = ebp[i]->lengthext;
+		} else {
+			if (ebp[i]->kinp->lengthset) {
+				leng = ebp[i]->kinp->lengthext;
+			} else {
+				fprintf(stderr, "WARNING: no branchlength set, defaults to 0.0!!!");
+				leng = 0.0;
+			}
+		}
+		alllen += leng;
+		if (leng < MINARC) leng = MINARC;
+		if (leng > MAXARC) leng = MAXARC;
+		if (clockmode) { /* clock */
+			ebp[i]->lengthc = leng;
+			ebp[i]->kinp->lengthc = leng;
+		} else { /* no clock */
+			ebp[i]->length = leng;
+			ebp[i]->kinp->length = leng;
+		}
+		Brnlength[i] = leng;
+	}
+	for (i = 0, j = numspc; i < numibrnch; i++, j++) {
+		if (ibp[i]->lengthset) {
+			leng = ibp[i]->lengthext;
+		} else {
+			if (ibp[i]->kinp->lengthset) {
+				leng = ibp[i]->kinp->lengthext;
+			} else {
+				fprintf(stderr, "WARNING: no branchlength set, defaults to 0.0!!!");
+				leng = 0.0;
+			}
+		}
+		alllen += leng;
+		if (leng < MINARC) leng = MINARC;
+		if (leng > MAXARC) leng = MAXARC;
+		if (clockmode) { /* clock */
+			ibp[i]->lengthc = leng;
+			ibp[i]->kinp->lengthc = leng;
+		} else { /* no clock */
+			ibp[i]->length = leng;
+			ibp[i]->kinp->length = leng;
+		}
+		Brnlength[j] = leng;
+	}
+} /* setextlength */
+
+
