@@ -2,16 +2,53 @@
  * util.c
  *
  *
- * Part of TREE-PUZZLE 5.0 (June 2000)
+ * Part of TREE-PUZZLE 5.2 (July 2004)
  *
- * (c) 1999-2000 by Heiko A. Schmidt, Korbinian Strimmer,
+ * (c) 2003-2004 by Heiko A. Schmidt, Korbinian Strimmer, and Arndt von Haeseler
+ * (c) 1999-2003 by Heiko A. Schmidt, Korbinian Strimmer,
  *                  M. Vingron, and Arndt von Haeseler
  * (c) 1995-1999 by Korbinian Strimmer and Arndt von Haeseler
  *
  * All parts of the source except where indicated are distributed under
  * the GNU public licence.  See http://www.opensource.org for details.
+ *
+ * ($Id$)
+ *
  */
 
+#ifdef HAVE_CONFIG_H
+#  include <config.h>
+#endif
+
+/*
+ * Use Scalable Parallel Random Number Generator
+ * See sprng.cs.fsu.edu for details
+ * from National Center for Supercomputing Applications (NCSA)
+ * See www.ncsa.uiuc.edu
+ */
+#ifdef NOSPRNG
+#   ifdef SPRNG
+#      undef SPRNG
+#   endif
+#else
+#   define SPRNG 1
+#endif
+
+#ifdef SPRNG
+#  ifdef PARALLEL
+#    define USE_MPI
+#    ifndef EXTERN
+#      define EXTERN extern
+#    endif
+#    include "ppuzzle.h"
+#  endif
+#  if 0
+#    define SIMPLE_SPRNG
+#  endif
+#  include "sprng/sprng.h"
+#endif
+
+
 
 #include "util.h"
 
@@ -28,23 +65,27 @@
 	extern int PP_Myid;
         long int PP_randn;
         long int PP_rand;
+	void PP_Finalize();
 #endif
 
 
-/*
+/**********************************
  * memory allocation error handler
  */
 
 void maerror(char *message)
 {
-	FPRINTF(STDOUTFILE "\n\n\nUnable to proceed (lack of memory: %s)\n\n", message);
-	FPRINTF(STDOUTFILE "Hint for Macintosh users:\n");
-	FPRINTF(STDOUTFILE "Use the <Get Info> command of the Finder to increase the memory partition!\n\n");
+	fprintf(STDOUT, "\n\n\nUnable to proceed (lack of memory: %s)\n\n", message);
+	fprintf(STDOUT, "Hint for Macintosh users:\n");
+	fprintf(STDOUT, "Use the <Get Info> command of the Finder to increase the memory partition!\n\n");
+#	if PARALLEL
+		PP_Finalize();
+#	endif
 	exit(1);
-}
+} /* maerror */
 
 
-/*
+/******************************************************
  * memory allocate double vectors, matrices, and cubes
  */
 
@@ -52,40 +93,46 @@ dvector new_dvector(int n)
 {
 	dvector v;
 
-	v = (dvector) malloc((unsigned) (n * sizeof(double)));
+	v = (dvector) calloc((size_t) n, sizeof(double));
 	if (v == NULL) maerror("step 1 in new_dvector");
 
 	return v;
-}
+} /* new_dvector */
+
+
+/******************/
 
 dmatrix new_dmatrix(int nrow, int ncol)
 {
 	int i;
 	dmatrix m;
 
-	m = (dmatrix) malloc((unsigned) (nrow * sizeof(dvector)));
+	m = (dmatrix) calloc((size_t) nrow, sizeof(dvector));
 	if (m == NULL) maerror("step 1 in in new_dmatrix");
 
-	*m = (dvector) malloc((unsigned) (nrow * ncol * sizeof(double)));
+	*m = (dvector) calloc((size_t) (nrow * ncol), sizeof(double));
 	if (*m == NULL) maerror("step 2 in in new_dmatrix");
 
 	for (i = 1; i < nrow; i++) m[i] = m[i-1] + ncol;
 
 	return m;
-}
+} /* new_dmatrix */
+
+
+/******************/
 
 dcube new_dcube(int ntri, int nrow, int ncol)
 {
 	int i, j;
 	dcube c;
 
-	c = (dcube) malloc((unsigned) (ntri * sizeof(dmatrix)));
+	c = (dcube) calloc((size_t) ntri, sizeof(dmatrix));
 	if (c == NULL) maerror("step 1 in in new_dcube");
 
-	*c = (dmatrix) malloc((unsigned) (ntri * nrow * sizeof(dvector)));
+	*c = (dmatrix) calloc((size_t) (ntri * nrow), sizeof(dvector));
 	if (*c == NULL) maerror("step 2 in in new_dcube");
 
-	**c = (dvector) malloc((unsigned) (ntri * nrow * ncol * sizeof(double)));
+	**c = (dvector) calloc((size_t) (ntri * nrow * ncol), sizeof(double));
 	if (**c == NULL) maerror("step 3 in in new_dcube");
 
 	for (j = 1; j < nrow; j++) c[0][j] = c[0][j-1] + ncol;
@@ -97,28 +144,37 @@ dcube new_dcube(int ntri, int nrow, int ncol)
 	}
 
 	return c;
-}
+} /* new_dcube */
+
+
+/******************/
 
 void free_dvector(dvector v)
 {
 	free((double *) v);
-}
+} /* free_dvector */
+
+
+/******************/
 
 void free_dmatrix(dmatrix m)
 {
 	free((double *) *m);
 	free((double *) m);
-}
+} /* free_dmatrix */
+
+
+/******************/
 
 void free_dcube(dcube c)
 {
 	free((double *) **c);
 	free((double *) *c);
 	free((double *) c);
-}
+} /* free_dcube */
 
 
-/*
+/****************************************************
  * memory allocate char vectors, matrices, and cubes
  */
 
@@ -126,40 +182,46 @@ cvector new_cvector(int n)
 {
 	cvector v;
 
-	v = (cvector) malloc((unsigned)n * sizeof(char));
+	v = (cvector) calloc((size_t) n, sizeof(char));
 	if (v == NULL) maerror("step1 in new_cvector");
 
 	return v;
-}
+} /* new_cvector */
+
+
+/******************/
 
 cmatrix new_cmatrix(int nrow, int ncol)
 {
 	int i;
 	cmatrix m;
 
-	m = (cmatrix) malloc((unsigned) (nrow * sizeof(cvector)));
+	m = (cmatrix) calloc((size_t) nrow, sizeof(cvector));
 	if (m == NULL) maerror("step 1 in new_cmatrix");
 
-	*m = (cvector) malloc((unsigned) (nrow * ncol * sizeof(char)));
+	*m = (cvector) calloc((size_t) (nrow * ncol), sizeof(char));
 	if (*m == NULL) maerror("step 2 in new_cmatrix");
 
 	for (i = 1; i < nrow; i++) m[i] = m[i-1] + ncol;
 
 	return m;
-}
+} /* new_cmatrix */
+
+
+/******************/
 
 ccube new_ccube(int ntri, int nrow, int ncol)
 {
 	int i, j;
 	ccube c;
 
-	c = (ccube) malloc((unsigned) (ntri * sizeof(cmatrix)));
+	c = (ccube) calloc((size_t) ntri, sizeof(cmatrix));
 	if (c == NULL) maerror("step 1 in new_ccube");
 
-	*c = (cmatrix) malloc((unsigned) (ntri * nrow * sizeof(cvector)));
+	*c = (cmatrix) calloc((size_t) (ntri * nrow), sizeof(cvector));
 	if (*c == NULL) maerror("step 2 in new_ccube");
 
-	**c = (cvector) malloc((unsigned) (ntri * nrow * ncol * sizeof(char)));
+	**c = (cvector) calloc((unsigned) (ntri * nrow * ncol), sizeof(char));
 	if (**c == NULL) maerror("step 3 in new_ccube");
 
 	for (j = 1; j < nrow; j++) c[0][j] = c[0][j-1] + ncol;
@@ -171,28 +233,37 @@ ccube new_ccube(int ntri, int nrow, int ncol)
 	}
 
 	return c;
-}
+} /* new_ccube */
+
+
+/******************/
 
 void free_cvector(cvector v)
 {
 	free((char *) v);
-}
+} /* free_cvector */
+
+
+/******************/
 
 void free_cmatrix(cmatrix m)
 {
 	free((char *) *m);
 	free((char *) m);
-}
+} /* free_cmatrix */
+
+
+/******************/
 
 void free_ccube(ccube c)
 {
 	free((char *) **c);
 	free((char *) *c);
 	free((char *) c);
-}
+} /* free_ccube */
 
 
-/*
+/***************************************************
  * memory allocate int vectors, matrices, and cubes
  */
 
@@ -200,40 +271,46 @@ ivector new_ivector(int n)
 {
 	ivector v;
 
-	v = (ivector) malloc((unsigned) (n * sizeof(int)));
+	v = (ivector) calloc((size_t) n, sizeof(int));
 	if (v == NULL) maerror("step 1 in new_ivector");
 
 	return v;
-}
+} /* new_ivector */
+
+
+/******************/
 
 imatrix new_imatrix(int nrow, int ncol)
 {
 	int i;
 	imatrix m;
 
-	m = (imatrix) malloc((unsigned) (nrow * sizeof(ivector)));
+	m = (imatrix) calloc((size_t) nrow, sizeof(ivector));
 	if (m == NULL) maerror("step 1 in new_imatrix");
 
-	*m = (ivector) malloc((unsigned) (nrow * ncol * sizeof(int)));
+	*m = (ivector) calloc((size_t) (nrow * ncol), sizeof(int));
 	if (*m == NULL) maerror("step 2 in new_imatrix");
 
 	for (i = 1; i < nrow; i++) m[i] = m[i-1] + ncol;
 
 	return m;
-}
+} /* new_imatrix */
+
+
+/******************/
 
 icube new_icube(int ntri, int nrow, int ncol)
 {
 	int i, j;
 	icube c;
 
-	c = (icube) malloc((unsigned) (ntri * sizeof(imatrix)));
+	c = (icube) calloc((size_t) ntri, sizeof(imatrix));
 	if (c == NULL) maerror("step 1 in new_icube");
 
-	*c = (imatrix) malloc((unsigned) (ntri * nrow * sizeof(ivector)));
+	*c = (imatrix) calloc((size_t) (ntri * nrow), sizeof(ivector));
 	if (*c == NULL) maerror("step 2 in new_icube");
 
-	**c = (ivector) malloc((unsigned) (ntri * nrow * ncol * sizeof(int)));
+	**c = (ivector) calloc((size_t) (ntri * nrow * ncol), sizeof(int));
 	if (**c == NULL) maerror("step 3 in new_icube");
 
 	for (j = 1; j < nrow; j++) c[0][j] = c[0][j-1] + ncol;
@@ -245,28 +322,37 @@ icube new_icube(int ntri, int nrow, int ncol)
 	}
 	
 	return c;
-}
+} /* new_icube */
+
+
+/******************/
 
 void free_ivector(ivector v)
 {
 	free((int *) v);
-}
+} /* free_ivector */
+
+
+/******************/
 
 void free_imatrix(imatrix m)
 {
 	free((int *) *m);
 	free((int *) m);
-}
+} /* free_imatrix */
+
+
+/******************/
 
 void free_icube(icube c)
 {
 	free((int *) **c);
 	free((int *) *c);
 	free((int *) c);
-}
+} /* free_icube */
 
 
-/*
+/***************************************************
  * memory allocate uli vectors, matrices, and cubes
  */
 
@@ -274,40 +360,46 @@ ulivector new_ulivector(int n)
 {
 	ulivector v;
 
-	v = (ulivector) malloc((unsigned) (n * sizeof(uli)));
+	v = (ulivector) calloc((size_t) n, sizeof(uli));
 	if (v == NULL) maerror("step 1 in new_ulivector");
 
 	return v;
-}
+} /* new_ulivector */
+
+
+/******************/
 
 ulimatrix new_ulimatrix(int nrow, int ncol)
 {
 	int i;
 	ulimatrix m;
 
-	m = (ulimatrix) malloc((unsigned) (nrow * sizeof(ulivector)));
+	m = (ulimatrix) calloc((size_t) nrow, sizeof(ulivector));
 	if (m == NULL) maerror("step 1 in new_ulimatrix");
 
-	*m = (ulivector) malloc((unsigned) (nrow * ncol * sizeof(uli)));
+	*m = (ulivector) calloc((size_t) (nrow * ncol), sizeof(uli));
 	if (*m == NULL) maerror("step 2 in new_ulimatrix");
 
 	for (i = 1; i < nrow; i++) m[i] = m[i-1] + ncol;
 
 	return m;
-}
+} /* new_ulimatrix */
+
+
+/******************/
 
 ulicube new_ulicube(int ntri, int nrow, int ncol)
 {
 	int i, j;
 	ulicube c;
 
-	c = (ulicube) malloc((unsigned) (ntri * sizeof(ulimatrix)));
+	c = (ulicube) calloc((size_t) ntri, sizeof(ulimatrix));
 	if (c == NULL) maerror("step 1 in new_ulicube");
 
-	*c = (ulimatrix) malloc((unsigned) (ntri * nrow * sizeof(ulivector)));
+	*c = (ulimatrix) calloc((size_t) (ntri * nrow), sizeof(ulivector));
 	if (*c == NULL) maerror("step 2 in new_ulicube");
 
-	**c = (ulivector) malloc((unsigned) (ntri * nrow * ncol * sizeof(uli)));
+	**c = (ulivector) calloc((size_t) (ntri * nrow * ncol), sizeof(uli));
 	if (**c == NULL) maerror("step 3 in new_ulicube");
 
 	for (j = 1; j < nrow; j++) c[0][j] = c[0][j-1] + ncol;
@@ -319,31 +411,86 @@ ulicube new_ulicube(int ntri, int nrow, int ncol)
 	}
 	
 	return c;
-}
+} /* new_ulicube */
+
+
+/******************/
 
 void free_ulivector(ulivector v)
 {
 	free((uli *) v);
-}
+} /* free_ulivector */
+
+
+/******************/
 
 void free_ulimatrix(ulimatrix m)
 {
 	free((uli *) *m);
 	free((uli *) m);
-}
+} /* free_ulimatrix */
+
+
+/******************/
 
 void free_ulicube(ulicube c)
 {
 	free((uli *) **c);
 	free((uli *) *c);
 	free((uli *) c);
-}
+} /* free_ulicube */
 
 
+/******************************************************************************/
+/* check time                                                                 */
+/******************************************************************************/
+#define TIMECHECK_INTERVALL 300
+/* check timer and generate message roughly every TIMECHECK_INTERVALL seconds */
+void checktime(time_t *starttime, 	/* starttime of overall part   */
+               time_t *lasttime,	/* lasttime of print out       */
+               time_t *nowtime,		/* current time (taken inside) */
+               uli     done,		/* number of tasks done        */
+               uli     sumtodo,		/* sum of task todo            */
+               int    *mflag)		/* LF needed to end line?      */
+{
+	double tc2, mintogo, minutes, hours;
+
+	/* check timer */
+	time(nowtime);
+	if ( (*nowtime - *lasttime) > TIMECHECK_INTERVALL) {
+		/* every TIMECHECK_INTERVAL seconds */
+		/* percentage of completed tasks */
+		if (*mflag == 0) {
+			fprintf(STDOUT, "\n");
+			*mflag = 1;
+		}
+		tc2 = 100.0 * done/sumtodo;
+		mintogo = (100.0-tc2) *
+			(double) (*nowtime-*starttime)/60.0/tc2;
+		hours = floor(mintogo/60.0);
+		minutes = mintogo - 60.0*hours;
+		fprintf(STDOUT, "%2.2f%%", tc2);
+		fprintf(STDOUT, " completed (remaining");
+		fprintf(STDOUT, " time: %.0f", hours);
+		fprintf(STDOUT, " hours %.0f", minutes);
+		fprintf(STDOUT, " minutes)\n");
+		fflush(STDOUT);
+		*lasttime = *nowtime;
+	} 
+	/* else {fprintf(stdout,"t%ld,", *nowtime - *lasttime); fflush(stdout);} */
+
+} /* checktime */
+
+
+#ifndef SPRNG
+
 /******************************************************************************/
 /* random numbers generator  (Numerical recipes)                              */
 /******************************************************************************/
 
+/* variable */
+long idum;
+
 /* definitions */
 #define IM1 2147483563
 #define IM2 2147483399
@@ -360,9 +507,6 @@ void free_ulicube(ulicube c)
 #define EPS 1.2e-7
 #define RNMX (1.0-EPS)
 
-/* variable */
-long idum;
-
 double randomunitintervall()
 /* Long period (> 2e18) random number generator. Returns a uniform random
    deviate between 0.0 and 1.0 (exclusive of endpoint values).
@@ -411,7 +555,7 @@ double randomunitintervall()
 		return RNMX;
 	else
 		return temp;
-}
+} /* randomunitintervall */
 
 #undef IM1
 #undef IM2
@@ -428,81 +572,163 @@ double randomunitintervall()
 #undef EPS
 #undef RNMX
 
-int initrandom(int seed)
+
+int initrandom(int seed)   /* RAND4 */
 {
    srand((unsigned) time(NULL));
    if (seed < 0) 
 	seed = rand();
    idum=-(long) seed;
-#  ifdef PARALLEL
-	{
-	int n;
-	for (n=0; n<PP_Myid; n++)
+	     fflush(stderr);
+#  ifndef PARALLEL
+	   fprintf(stderr,"Using RAND4 Random Number Generator\n");
+#	   ifdef RANDVERBOSE1
+	      fprintf(stderr, "!!! random seed set to %d !!!\n", seed);
+#	   endif  /* RANDVERBOSE1 */
+#  else /* PARALLEL */
+	   {
+	   int n;
+	   for (n=0; n<PP_Myid; n++)
 		(void) randomunitintervall();
-#       ifdef PVERBOSE1
-  	   fprintf(stderr, "(%2d) !!! random seed set to %d, %dx drawn !!!\n", PP_Myid, seed, n);
-#       endif
-	}
-#  else
-#       ifdef PVERBOSE1
-	   fprintf(stderr, "!!! random seed set to %d !!!\n", seed);
-#       endif
+#	   ifdef RANDVERBOSE1
+	      fprintf(stderr, "(%2d) !!! random seed set to %d, %dx drawn !!!\n", PP_Myid, seed, n);
+#	   endif  /* RANDVERBOSE1 */
+	   }
 #  endif
    return (seed);
 }  /* initrandom */ 
 
+/******************/
+
+#else /* USE_SPRNG */
+
+/******************/
+
+int *randstream;
+
+int initrandom(int seed)
+{
+   srand((unsigned) time(NULL));
+   if (seed < 0) 
+	seed = make_sprng_seed();
+#  ifndef PARALLEL
+	   fprintf(stderr,"Using SPRNG -- Scalable Parallel Random Number Generator\n");
+	   fflush(stderr);
+	   randstream = init_sprng(0,1,seed,SPRNG_DEFAULT); /*init stream*/
+
+#	   ifdef RANDVERBOSE1
+	      fprintf(stderr, "!!! random seed set to %d !!!\n", seed);
+	      printf(" Printing information about new stream\n");
+	      print_sprng(randstream);
+#	   endif /* RANDVERBOSE 1 */
+#  else /* PARALLEL */
+	   if (PP_IamMaster) {
+	     fprintf(stderr,"Using SPRNG -- Scalable Parallel Random Number Generator\n");
+	     fflush(stderr);
+	   }
+	   /* MPI_Bcast(&seed, 1, MPI_UNSIGNED, PP_MyMaster, MPI_COMM_WORLD); */
+	   randstream = init_sprng(PP_Myid,PP_NumProcs,seed,SPRNG_DEFAULT); /*initialize stream*/
+#	   ifdef RANDVERBOSE1
+	      fprintf(stderr, "(%2d) !!! random seed set to %d !!!\n", PP_Myid, seed);
+	      printf(" Printing information about new stream\n");
+	      print_sprng(randstream);
+#	   endif  /* RANDVERBOSE1 */
+#  endif /* PARALLEL */
+   return (seed);
+}  /* initrandom */ 
+
+
+#endif /* USE_SPRNG */
+
+/******************/
+
 
 /* returns a random integer in the range [0; n - 1] */
 int randominteger(int n)
 {
-	int t;
 #  ifndef FIXEDINTRAND
 #	ifndef PARALLEL
+#	   ifdef SPRNG
+		return (int) floor(sprng(randstream)*n);
+#	   else /* NO_SPRNG */
+		int t;
 		t = (int) floor(randomunitintervall()*n);
 		return t;
-#	else
+#	   endif /* NO_SPRNG */
+#	else /* NOT PARALLEL */
+#	   ifdef SPRNG
+		return (int) floor(sprng(randstream)*n);
+#	   else /* NO_SPRNG */
 		int m;
 		for (m=1; m<PP_NumProcs; m++)
 			(void) randomunitintervall();
 		PP_randn+=(m-1); PP_rand++;
 		return (int) floor(randomunitintervall()*n);
-#	endif
-#  else
+#	   endif /* NO_SPRNG */
+#	endif /* NOT PARALLEL */
+#  else /* FIXEDINTRAND */
 	fprintf(stderr, "!!! fixed \"random\" integers for testing purposes !!!\n");
 	return (int)0;
 #  endif /* FIXEDINTRAND */
-}
+} /* randominteger */
+
+
+/******************/
 
 /* Draw s numbers from the set 0,1,2,..,t-1 and put them
    into slist (every number can be drawn only one time) */
 void chooser(int t, int s, ivector slist)
 {
-	int i, j, k, l;
+	int i;
+	int randidx;
 	ivector isfree;
 
 	isfree = new_ivector(t);
-	for (i = 0; i < t; i++) isfree[i] = TRUE;
-	for (i = 0; i < s; i++) {
-		/* random integer in the range [0, t-1-i] */
-		j = randominteger(t-i);
-		k = -1;
-		l = -1;
-		do {
-			k++;
-			if (isfree[k] == TRUE) l++; 
-		} while ( l != j);
-		slist[i] = k;
-		isfree[k] = FALSE;
+	for (i = 0; i < t; i++) isfree[i] = i;
+
+	for (i = t; i > t-s; i--) {
+		/* random integer in the range [0, still to do = i-1] */
+		randidx = randominteger(i);
+
+		/* integer from array of free ints to slist */
+		slist[t - i] = isfree[randidx];
+
+		/* reduce list by moving last free to last chosen */
+		isfree[randidx] = isfree[i-1];
 	}
 	free_ivector(isfree);
-}
+} /* chooser */
+
+
+
+/******************/
+
+/* Draw s numbers from the set 0,1,2,..,t-1 and put them
+   into slist (every number can be drawn only one time) */
+void fixedchooser(int t, int s, ivector slist)
+{
+	int i;
+
+	for (i = 0; i < s; i++) {
+		/* random integer in the range [0, s-1] */
+		slist[i] = i;
+	}
+} /* fixedchooser */
+
+
+/******************************************************************************/
+/* Saver variants of ANSI C routines                                          */
+/******************************************************************************/
 
 /* a realloc function that works also on non-ANSI compilers */ 
 void *myrealloc(void *p, size_t s)
 {
 	if (p == NULL) return malloc(s);
 	else return realloc(p, s);
-}
+} /* myrealloc */
+
+
+/******************/
 
 /* safer variant of gets */
 /* Reads characters from stdin until a newline character or EOF
@@ -517,7 +743,9 @@ cvector mygets()
 	
 	n = 0;
 	c = getchar();
-	while (c != '\n' && c != '\r' && n < 99 && c != EOF && !ferror(stdin))
+	while ((c != '\n') && (c != '\r') 
+               && (n < 99) && (c != EOF) 
+               && !ferror(stdin))
 	{
 		str[n] = (char) c;
 		c = getchar();
@@ -533,7 +761,7 @@ cvector mygets()
 	}
 	
 	return str;
-}
+} /* mygets */
 
 
 
@@ -541,7 +769,7 @@ cvector mygets()
 /* minimization of a function by Brents method (Numerical Recipes)            */
 /******************************************************************************/
 
-double brent(double, double, double, double (*f )(double ), double, double *, double *, double, double, double);
+double brent(double, double, double, double (*f )(double ), double, double *, double *, double, double, double, int *);
 
 
 #define ITMAX 100
@@ -554,18 +782,32 @@ double brent(double, double, double, double (*f )(double ), double, double *, do
 #define SIGN(a,b) ((b) >= 0.0 ? fabs(a) : -fabs(a))
 
 /* Brents method in one dimension */
-double brent(double ax, double bx, double cx, double (*f)(double), double tol,
-	double *foptx, double *f2optx, double fax, double fbx, double fcx)
+double brent(
+	double ax, 		/* three x-values ax < bx < cx */
+	double bx, 		/* or             ax > bx > cx */
+	double cx, 
+	double (*f)(double), 	/* function to minimize, f(bx)=min(f(ax),f(bx),f(cx)) */
+	double tol,		/* describes the precision */
+	double *foptx, 
+	double *f2optx, 
+	double fax, 		/* f(ax), already precomputed outside of brent() */
+	double fbx, 		/* f(bx), already precomputed outside of brent() */
+	double fcx,		/* f(cx), already precomputed outside of brent() */
+	int    *converged)	/* not converged error flag */
 {
 	int iter;
 	double a,b,d=0,etemp,fu,fv,fw,fx,p,q,r,tol1,tol2,u,v,w,x,xm;
 	double xw,wv,vx;
 	double e=0.0;
 
-	a=(ax < cx ? ax : cx);
+	*converged=TRUE;
+
+	a=(ax < cx ? ax : cx);	/* is the order correct? If not, exchange ax and cx */
 	b=(ax > cx ? ax : cx);
-	x=bx;
+
+	x=bx;			/* store 'middle' x, and y=fx */
 	fx=fbx;
+
 	if (fax < fcx) {
 		w=ax;
 		fw=fax;
@@ -577,19 +819,31 @@ double brent(double ax, double bx, double cx, double (*f)(double), double tol,
 		v=ax;
 		fv=fax;	
 	}
-	for (iter=1;iter<=ITMAX;iter++) {
+
+	for (iter=1;iter<=ITMAX;iter++) {		/* iteration loop */
 		xm=0.5*(a+b);
 		tol2=2.0*(tol1=tol*fabs(x)+ZEPS);
-		if (fabs(x-xm) <= (tol2-0.5*(b-a))) {
-			*foptx = fx;
+		/* is precision OK? I.e., are a and b close enough. */
+		if (fabs(x-xm) <= (tol2-0.5*(b-a))) { /* yes -> finish */
+			*foptx = fx;		/* return current f(x) */
+			if ((x==w) && (x==v)) {  
+				/* to avoid FPE errors by zero-division */
+				v=x-ZEPS;
+				w=x+ZEPS;
+				fv=(*f)(v);
+				fw=(*f)(w);
+			}
 			xw = x-w;
 			wv = w-v;
 			vx = v-x;
 			*f2optx = 2.0*(fv*xw + fx*wv + fw*vx)/
-				(v*v*xw + x*x*wv + w*w*vx);
-			return x;
+				(v*v*xw + x*x*wv + w*w*vx);	/* return S.E. of f(x) */
+#if 0
+fprintf(stderr, "XXXX: x=%f f(x)=%f=%f S.E.=%f   w=%f v=%f\n", x, *foptx, fx, *f2optx, w, v);
+#endif
+			return x;	/* return minimum x-value */
 		}
-		if (fabs(e) > tol1) {
+		if (fabs(e) > tol1) {		/* compute a parabolic fit */
 			r=(x-w)*(fx-fv);
 			q=(x-v)*(fx-fw);
 			p=(x-v)*q-(x-w)*r;
@@ -598,22 +852,28 @@ double brent(double ax, double bx, double cx, double (*f)(double), double tol,
 			q=fabs(q);
 			etemp=e;
 			e=d;
+			/* determine acceptability of the parabolic fit */
 			if (fabs(p) >= fabs(0.5*q*etemp) || p <= q*(a-x) || p >= q*(b-x))
+				/* compute golden section of the larger segment max(a-x,b-x) */
 				d=CGOLD*(e=(x >= xm ? a-x : b-x));
 			else {
+				/* accept the parabolic step */
 				d=p/q;
 				u=x+d;
 				if (u-a < tol2 || b-u < tol2)
 					d=SIGN(tol1,xm-x);
 			}
 		} else {
+			/* compute golden section of the larger segment max(a-x,b-x) */
 			d=CGOLD*(e=(x >= xm ? a-x : b-x));
 		}
 		u=(fabs(d) >= tol1 ? x+d : x+SIGN(tol1,d));
 		fu=(*f)(u);
-		if (fu <= fx) {
+
+		/* new u better than old x ? */
+		if (fu <= fx) { /* yes -> accept it */
 			if (u >= x) a=x; else b=x;
-			SHFT(v,w,x,u)
+			SHFT(v,w,x,u)		/* ...and use the new values */
 			SHFT(fv,fw,fx,fu)
 		} else {
 			if (u < x) a=u; else b=u;
@@ -627,15 +887,28 @@ double brent(double ax, double bx, double cx, double (*f)(double), double tol,
 				fv=fu;
 			}
 		}
+	} /* end loop */
+
+	/* the function should never get here, i.e., the search did not converge within ITMAX iterations */
+	*converged=FALSE;
+
+	*foptx = fx;	/* return f(x) */
+	if ((x==w) && (x==v)) {  /* regularize to avoid FPE errors by zero-division */
+		v=x-ZEPS;
+		w=x+ZEPS;
+		fv=(*f)(v);
+		fw=(*f)(w);
 	}
-	*foptx = fx;
 	xw = x-w;
 	wv = w-v;
 	vx = v-x;
 	*f2optx = 2.0*(fv*xw + fx*wv + fw*vx)/
-		(v*v*xw + x*x*wv + w*w*vx);
-	return x;
-}
+		(v*v*xw + x*x*wv + w*w*vx);	/* return S.E. of f(x) */
+#if 0
+fprintf(stderr, "YYYY: x=%f f(x)=%f=%f S.E.=%f   w=%f v=%f\n", x, *foptx, fx, *f2optx, w, v);
+#endif
+	return x;	/* return minimum x-value */
+} /* brent */
 #undef ITMAX
 #undef CGOLD
 #undef ZEPS
@@ -645,16 +918,21 @@ double brent(double ax, double bx, double cx, double (*f)(double), double tol,
 #undef GLIMIT
 #undef TINY
 
+
+/******************/
+
 /* one-dimensional minimization - as input a lower and an upper limit and a trial
   value for the minimum is needed: xmin < xguess < xmax
   the function and a fractional tolerance has to be specified
   onedimenmin returns the optimal x value and the value of the function
   and its second derivative at this point
   */
-double onedimenmin(double xmin, double xguess, double xmax, double (*f)(double),
+double onedimenmin(
+	double xmin, double xguess, double xmax, double (*f)(double),
 	double tol, double *fx, double *f2x)
 {
 	double eps, optx, ax, bx, cx, fa, fb, fc;
+	int    converged;	/* not converged error flag */
 		
 	/* first attempt to bracketize minimum */
 	eps = xguess*tol*50.0;
@@ -673,12 +951,15 @@ double onedimenmin(double xmin, double xguess, double xmax, double (*f)(double),
 	if ((fa < fb) || (fc < fb)) {
 		if (ax != xmin) fa = (*f)(xmin);
 		if (cx != xmax) fc = (*f)(xmax);
-		optx = brent(xmin, xguess, xmax, f, tol, fx, f2x, fa, fb, fc);
+		optx = brent(xmin, xguess, xmax, f, tol, fx, f2x, fa, fb, fc, &converged);
 	} else
-		optx = brent(ax, bx, cx, f, tol, fx, f2x, fa, fb, fc);
+		optx = brent(ax, bx, cx, f, tol, fx, f2x, fa, fb, fc, &converged);
 
 	return optx; /* return optimal x */
-}
+} /* onedimenmin */
+
+
+/******************/
 
 /* two-dimensional minimization with borders and calculations of standard errors */
 /* we optimize along basis vectors - not very optimal but it seems to work well */
@@ -744,5 +1025,5 @@ void twodimenmin(double tol,
 	} while (it != MAXITS && change);
 
 	return;
-}
+} /* twodimenmin */
 
